// Morgan Elder
// CMPSCI 4760

// Project 2
// master.c
// [ENTER SUMMARY OF PROGRAM HERE]

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <stdbool.h>
#include <sys/wait.h>
#include <time.h>
#include <signal.h>
#include <stdarg.h>
#include "config.h"

// shared resources
typedef struct sharedData
{
  int *number;
  bool *choosing;
} shm_data;

shm_data *shared_mem;

int *childPidList;
key_t key = 5678;
int n;

static int isNumber(char *num_string);
static void doConcurrentProcs(int n, int timeout);
static int createSeg();
static int initSharedData(int id, int n);
static void detachSharedData();
static void removeSharedData(int id);
static void doThread(int i);
static void lock(int n, int i);
static int getMax(int *arr, int size);
static void unlock(int i);
static int getRandomInt(int upper, int lower);
static void sigHandler(int signum);
static void writeLog(int entry, int procNum, int que);

int main(int argc, char *argv[])
{
  int opt;
  int timeout = MAX_TIMEOUT;
  n = MAX_PROCESSES;

  srand(time(0));

  if ((opt = getopt(argc, argv, "t:")) != -1)
  {
    switch (opt)
    {
    case 't':
      if (optind != 3)
      {
        fprintf(stderr, "%s: error: Option %c must be the first argument\n", argv[0], opt);
        return 1;
      }
      else if (isNumber(optarg) == -1)
      {
        fprintf(stderr, "%s: error: arg for option %c must be integer value\n", argv[0], opt);
        return 1;
      }
      timeout = atoi(optarg);
      if (timeout > MAX_TIMEOUT)
      {
        fprintf(stderr, "%s: error: arg for option %c cannot exceed %d\n", argv[0], opt, MAX_TIMEOUT);
        return 1;
      }
      else if (timeout < MIN_TIMEOUT)
      {
        fprintf(stderr, "%s: error: arg for option %c cannot be negative\n", argv[0], opt);
        return 1;
      }

      break;
    case '?':
      fprintf(stderr, "%s: error: Uknown option %c\n", argv[0], optopt);
      return 1;
    case ':':
      fprintf(stderr, "%s: error: Missing arg for option %c\n", argv[0], opt);
      return 1;
    default:
      abort();
    }
  }

  if (optind > argc - 1)
  {
    fprintf(stderr, "%s: error: Missing arg specifying number of processes\n", argv[0]);
    return 1;
  }
  else if (isNumber(argv[argc - 1]) == -1)
  {
    fprintf(stderr, "%s: error: Arg specifying number of processes must be an number\n", argv[0]);
    return 1;
  }
  else
  {
    n = atoi(argv[argc - 1]);
    if (n > MAX_PROCESSES)
    {
      fprintf(stderr, "%s: error: Arg specifying number of processes cannot exceed %d\n", argv[0], MAX_PROCESSES);
      return 1;
    }
    else if (n < MIN_PROCESSES)
    {
      fprintf(stderr, "%s: error: Arg specifying number of processes cannot be less than %d\n", argv[0], MIN_PROCESSES);
      return 1;
    }
    doConcurrentProcs(n, timeout);
  }

  return 0;
}

static int isNumber(char *num_string)
{
  int i;
  for (i = num_string[0] == '-' ? 1 : 0; num_string[i] != '\0'; i++)
  {
    if (isdigit(num_string[i]))
    {
      continue;
    }
    else
    {
      return -1;
    }
  }
  return 0;
}

static void doConcurrentProcs(int n, int timeout)
{

  int shmid;
  int pid;

  childPidList = (int *)malloc(n * sizeof(int));

  // allocate shared memory
  // create segment with shmget(key, size, int shmflag)
  shmid = createSeg();

  // attach shared memory
  initSharedData(shmid, n);

  int i;
  for (i = 0; i < n; i++)
  {
    pid = fork(); // fork returns 0 for child process and greater than 0 for parent
    if (pid < 0)
    {
      perror("fork");
    }
    else if (pid == 0)
    {
      doThread(i + 1);
      exit(0);
    }
    else
    {
      childPidList[i] = pid;
    }
  }

  if (pid > 0)
  {
    signal(SIGINT, sigHandler);
    signal(SIGALRM, sigHandler);
    alarm(timeout);

    int activeChildProcCount = n;
    while (activeChildProcCount > 0)
    {
      if (wait(NULL) == -1)
      {
        perror("wait");
      }
      activeChildProcCount--;
    }
    // only parent process can free shared memory resources
    // detach shared memory
    detachSharedData();

    // remove shared mem
    removeSharedData(shmid);

    free(childPidList);
  }
}

static int createSeg()
{
  int shmid;
  // key_t key = 5678;

  shmid = shmget(key, sizeof(shm_data), IPC_CREAT | 0600);
  if (shmid < 0)
  {
    perror("shmget");
    exit(EXIT_FAILURE);
  }
  return shmid;
}

static int initSharedData(int id, int n)
{
  // attach shared memory
  shared_mem = (shm_data *)shmat(id, NULL, 0);
  if (!shared_mem)
  {
    perror("shmat");
    exit(EXIT_FAILURE);
  }
  shared_mem->choosing = (bool *)malloc(n * sizeof(bool));
  shared_mem->number = (int *)malloc(n * sizeof(int));

  int i;
  for (i = 0; i < n; i++)
  {
    shared_mem->choosing[i] = false;
    shared_mem->number[i] = 0;
  }
  return 0;
}

static void detachSharedData()
{
  free(shared_mem->choosing);
  free(shared_mem->number);
  if (shmdt(shared_mem) == -1)
  {
    perror("shmdt");
    exit(EXIT_FAILURE);
  }
}

static void removeSharedData(int id)
{
  if (shmctl(id, IPC_RMID, NULL) == -1)
  {
    perror("shmctl");
    exit(EXIT_FAILURE);
  }
}

static void doThread(int i)
{
  int randMin = 1;
  int randMax = 5;
  char arg1[10];
  char arg2[10];

  int iter;
  for (iter = 0; iter < 5; iter++)
  {
    lock(n, i); // entering critical section
    writeLog(3, i, shared_mem->number[i]);
    // critical section
    sleep(getRandomInt(randMax, randMin));
    sprintf(arg1, "%d", i);
    sprintf(arg2, "%d", shared_mem->number[i]);
    if (fork() == 0)
    {
      execl("./slave", arg1, arg2, NULL);
      printf("%d\n", i);
    }
    else
    {
      wait(NULL);
    }
    sleep(getRandomInt(randMax, randMin));
    writeLog(4, i, shared_mem->number[i]);
    unlock(i); // exiting critical section
  }
}

static void lock(int n, int i)
{
  shared_mem->choosing[i] = true;
  shared_mem->number[i] = getMax(shared_mem->number, n) + 1;
  shared_mem->choosing[i] = false;

  // entering critical section
  int j;
  for (int j = 0; j < n; j++)
  {
    while (shared_mem->choosing[j])
    {
    }

    while (shared_mem->number[j] != 0 &&
           (shared_mem->number[i] < shared_mem->number[i] ||
            (shared_mem->number[j] == shared_mem->number[i] && j < i)))
    {
    }
  }
}

static int getMax(int *arr, int size)
{
  int max = arr[0];

  int i;
  for (i = 1; i < size; i++)
  {
    if (max < arr[i])
    {
      max = arr[i];
    }
  }
  return max;
}

static void unlock(int i)
{
  shared_mem->number[i] = 0;
}

static int getRandomInt(int upper, int lower)
{
  return (rand() % (upper - lower + 1)) + lower;
}

// sigHandler handles the signals generated by a timeout alarm and
// user interrupts via pressing CTRL+C
// The method terminates child programs and frees shared memory
static void sigHandler(int signum)
{

  // kill child processes
  int i;
  for (i = 0; i < n; i++)
  {
    kill(childPidList[i], SIGKILL);
  }

  // write to log
  if (signum == SIGALRM)
  {
    printf("Caught signal %d: Timeout initiated\n", signum);
    writeLog(1, 0, 0);
  }
  else if (signum == SIGINT)
  {
    printf("Caught signal %d: User interrupt\n", signum);
    writeLog(2, 0, 0);
  }

  // free shared memory pointers and
  // deallocate shared memory
  detachSharedData();
  // remove shared memory
  removeSharedData(shmget(key, 0, 0));

  free(childPidList);
  exit(1);
}

// writeLog records events that occur in master by parent and child processes
// 1. timer runs up
// 2. user sends interrupt signal with CTRL+C
// 3. Child process is entering critical section
// 4. Child process is exiting critical section
static void writeLog(int entry, int procNum, int queue)
{
  FILE *logFile;
  char *message;
  char *fileName = "logfile";

  time_t rawTime;
  char currentTime[12];

  time(&rawTime);
  strftime(currentTime, 12, "%H:%M:%S", localtime(&rawTime));

  switch (entry)
  {
  case 1:
    asprintf(&fileName, "%s", fileName);
    asprintf(&message, "%s Timeout limit reach. Child Processes terminated. Shared memory deallocated.\n", currentTime);
    break;
  case 2:
    asprintf(&fileName, "%s", fileName);
    asprintf(&message, "%s Interrupt signal received. Child Processes terminated. Shared memory deallocated.\n", currentTime);
    break;
  case 3:
    asprintf(&fileName, "%s.%d", fileName, procNum);
    asprintf(&message, "%s Queue %d Entering critical by process number %d\n", currentTime, queue, procNum);
    break;
  case 4:
    asprintf(&fileName, "%s.%d", fileName, procNum);
    asprintf(&message, "%s Queue %d Exiting critical by process number %d\n", currentTime, queue, procNum);
    break;
  }

  logFile = fopen(fileName, "a");
  fputs(message, logFile);
  free(message);
  free(fileName);
  fclose(logFile);
}